//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

//  Files-11 Structure level 2 

#ifndef __BITMAP_H__
#define __BITMAP_H__

#include "definitions.h"
#include "f11asrt.h"
#include <stdio.h>

class bitmap {

private:
    ods_dword_t*	array;
    size_t		bit_count;
    size_t		ff_bit;

    virtual dword_t	free_value() = 0;
    virtual dword_t	busy_value() = 0;

public:
    bitmap() : array(NULL), bit_count(0), ff_bit(-1) {}

    bitmap(ods_dword_t* bmap, size_t cnt) : array(bmap), bit_count(cnt), ff_bit(-1) {}

    void init(ods_dword_t* bmap, size_t cnt, size_t start_bit = 0) {
	ASSERT(start_bit <= cnt);

	this->array = bmap; 
	this->bit_count = cnt;
	this->ff_bit = -1;
	update_ff_bit(start_bit);
    }

    size_t  get_bit_count() { 
	return this->bit_count; 
    }

    size_t  get_ff_bit() { 
	if (this->ff_bit = (size_t)-1) update_ff_bit();
	if (this->ff_bit >= this->bit_count) return -1;
	else {
	    // Make sure that ff bit value is set properly
	    ASSERT(get_bit(this->ff_bit) == this->free_value());
	    return this->ff_bit; 
	}
    }

protected:
    size_t  update_ff_bit(size_t start_bit = 0) {
	ASSERT(start_bit >= 0 && start_bit < bit_count);

	dword_t	v;
	size_t	i = start_bit / (sizeof(dword_t)*8);
	dword_t	j = start_bit % (sizeof(dword_t)*8);
	size_t	max_dword = (bit_count-1) / (sizeof(dword_t)*8) + 1;
	dword_t	val = free_value();

	for (; i < max_dword; i++)  {
	    for (; j < (sizeof(dword_t)*8); j++) {
		v = ((array[i] >> j) & 1) ^ val;
		if (!v) {
		    this->ff_bit = i * (sizeof(dword_t)*8) + j;
		    TRACE(("updating cached bmap's ff_bit value (ff_bit = %lu)\n", ff_bit));
		    return ff_bit;
		}
	    }
	    j = 0;
	}

	// Free bit is not found, return bit_count (max+1)
	this->ff_bit = bit_count;
	return ff_bit;
    }

    void    set_bit_value(size_t bit, dword_t val) {
	ASSERT(bit >= 0 && bit < bit_count);

	size_t	i = bit / (sizeof(dword_t)*8);
	dword_t	j = bit % (sizeof(dword_t)*8);
	dword_t	mask = 1L << j;

	val = val ? mask : 0;
	array[i] = (array[i] & ~mask) | val;

	// Update the bmap's ff_bit cached value
	// since we've modified the bmap contents
	update_ff_bit( (bit > this->ff_bit) ? this->ff_bit : bit);
    }

    dword_t get_bit(size_t bit) {
	ASSERT(bit >= 0 && bit < bit_count);

	size_t	i = bit / (sizeof(dword_t)*8);
	dword_t	j = bit % (sizeof(dword_t)*8);

	return ((array[i] >> j) & 1L);
    }

public:
    // Assign the 'busy' value to a bit
    void    alloc_bit(dword_t bit) { 
	set_bit_value(bit, busy_value()); 
    }
    
    // Assign the 'free' value to a bit
    void    release_bit(dword_t bit) { 
	set_bit_value(bit, free_value()); 
    }
};

#endif 	/* __BITMAP_H__	*/
