//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

// Files-11 Structure level 2 

#include <stdlib.h>
#include <string.h>

#include "definitions.h"
#include "f11asrt.h"
#include "devio.h"
#include "files11.h"
#include "f11liberr.h"
#include "f11file.h"
#include "f11vol.h"
#include "f11file_sys.h"


//  Read/write of file block(s).
//  This function works with virtual blocks
int f11_file::f11_read_write_file_blk(
				  dword_t   vbn,
				  dword_t   blkcount,
				  int	    write)
{
    lbn_t	lbn   = INVALID_LBN;
    dword_t	count = 0, j, blk_complete = 0;
    file_hdr	*f, *f1;
    ODS_FH2*	fh;
    ODS_FM2  	*rp;
    int 	shift = -1;
    void*	buf = p_content_window->p_content;
    lbn_t*	lbns = p_content_window->p_block_lbn;

    ASSERT(blkcount && buf);

    TRACE(("f11_read_write_file_blk: vbn 0%d count 0%d write=%d\n", vbn, blkcount, write));

    //  EOF check
    if(vbn >= get_fhdr()->get_eofblk()) {
	TRACE(("f11_read_write_file_blk: vbn >= EOF blk\n"));
	dev->set_error(F11_EOF);
	return -1;
    }

    //  search apropriate LBN
    f = f1 = file_header;
    j = 0;
    do {

	if(!f1) {
	    TRACE(("f11_read_write_file_blk: no more extensions\n"));
	    dev->set_error(F11_EOF);
	    return blk_complete;
	}

	f = f1;
	ASSERT(f);

	//  Go through all used retrieval pointers
	fh = &f->fhdr;
	rp = fh->get_mp_area();

	rp_iterator next_rp(rp, fh->fh2_b_map_inuse);
	while ((rp = next_rp++) != NULL)
	{
	    // Handle placement header
	    if(rp->rp_isfmt0()) {
		continue;
	    }

	    // Found ?
	    j += rp->rp_getcount();
	    if (j > vbn) {
		shift = rp->rp_getcount() - (j - vbn);
		TRACE(("f11_read_write_file_blk: appropriate extension found, shift = %d\n", shift));
		break;
	    }
	}

	// Next extension
	f1 = f->next;
	TRACE(("f11_read_write_file_blk: go to next extension\n"));

    } while(shift == -1);

    //  I/O
    while(blk_complete != blkcount) {
	ASSERT(f);

	rp_iterator next_rp(rp, fh->fh2_b_map_inuse);
	while ((rp = next_rp++) != NULL && blk_complete != blkcount)
	{
	    if(rp->rp_isfmt0())
		continue;
	    lbn = rp->rp_getlbn() + shift;
	    count = rp->rp_getcount() - shift;

	    ASSERT(count);

	    count = blkcount - blk_complete < count ? 
		blkcount - blk_complete : count;

	    TRACE(("READ from rp: block(s) lbn 0%d count 0%d, total rp_count 0%d\n",
		   lbn, count, rp->rp_getcount()));

	    if (write)
		j = dev->f11_write_logical_blk(lbn, count, buf);
	    else
		j = dev->f11_read_logical_blk(lbn, count, buf);

	    if(j != count * F11_BLOCK_SIZE) {
		if(dev->get_dev_error()) {
		    TRACE(("f11_read_write_file_blk: drive error detected\n"));
		    dev->set_error(F11_DRIVE_ERR);
		    return -1;
		}
		dev->set_error(F11_EOF);
		return blk_complete + (j / F11_BLOCK_SIZE);
	    }

	    p_content_window->start_vbn = vbn;
	    for (unsigned int n = 0; n < j/F11_BLOCK_SIZE; n++) {
		*(lbns++) = lbn++;
	    }

	    buf 	  = (byte_t*)buf + j;
	    blk_complete += j / F11_BLOCK_SIZE;
	    shift 	  = 0;
	}

	//  no more extensions ? 
	if(!f->next && blk_complete != blkcount) {
	    TRACE(("f11_read_write_file_blk: no more extensions\n"));
	    dev->set_error(F11_EOF);
	    break;
	}

	f = f->next;
	if (f) {
	    fh = &f->fhdr;
	    rp = fh->get_mp_area();
	}
    }

    return blk_complete;
}


int f11_file::delete_file()
{
    lbn_t	lbn   = INVALID_LBN;
    dword_t	count = 0;
    file_hdr	*f;
    ODS_FH2*	fh;
    ODS_FM2  	*rp;
    int		cluster_size;

    TRACE(("delete file content and all headers\n"));

    cluster_size = dev->hblk.hm2_w_cluster;

    // Search apropriate LBN
    f = file_header;
    while (f != NULL) {

	//  Go through all used retrieval pointers
	fh = &f->fhdr;
	rp = fh->get_mp_area();

	rp_iterator next_rp(rp, fh->fh2_b_map_inuse);
	while ((rp = next_rp++) != NULL)
	{

	    // Handle placement header
	    if(rp->rp_isfmt0()) {
		continue;
	    }

	    // For each found retrieval pointer,
	    // releaze the blocks that it specifies
	    lbn = rp->rp_getlbn();
	    count = rp->rp_getcount();

	    // Release each cluster in the volume bitmap file
	    for (dword_t i = 0; i < count; i += cluster_size) {
		dev->p_bitmap_sys->release_cluster((lbn + i) / cluster_size);
	    }
	}

	// Mark the file header itself as free for further use
	// in the index file
	dword_t fid_num = fh->fh2_w_fid.numb();
	dev->p_indexf_sys->release_file_index(fh->fh2_w_fid.numb());

	f = f->next;
    };

    TRACE(("delete file content: done\n"));

    return 0;
}

//  Get number of block(s) mapped by file retrieval pointers
dword_t	f11_file::get_rp_blk_count() const
{
    dword_t	blk_count = 0;
    
    // Search for apropriate LBN
    for (file_hdr* f = file_header; f != NULL; f = f->next) {
	blk_count += f->fhdr.get_rp_blk_count();
    }

    return blk_count;
}


//  Syncing file content and header
int f11_file::f11_sync()
{
    file_hdr	*fl, *fl1;
    if (this->flag_header_modified) {
	TRACE(("Syncing of file headers is required...\n"));

	// Sync list of file headers (primary and all extensions)
	fl = file_header;
	while(fl) {
	    fl1 = fl->next;

	    // Calculate new checksum for the file header block...
	    fl->fhdr.update_checksum();

	    // ...and write it back to the volume
	    TRACE(("File header syncing...\n"));
	    dev->f11_write_logical_blk(fl->fhdr_lbn, 1, &fl->fhdr);
	    fl  = fl1;
	}
	this->flag_header_modified = 0;
    }

    // Sync file content
    if (this->flag_content_modified) {
	TRACE(("Syncing of file content is required...\n"));
	if (f11_write_file_blk(p_content_window->start_vbn, this->p_content_window->block_count) != (int)this->p_content_window->block_count) {
	    TRACE(("File content syncing FAILED!\n"));
	    return -1;
	}
	TRACE(("File content syncing done.\n"));
	this->flag_content_modified = 0;
    }

    TRACE(("File syncing: done.\n"));
    return 0;
}
