//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

// Files-11 Structure level 2 

#ifndef __F11FILE_SYS_H__
#define __F11FILE_SYS_H__

#include "f11file.h"
#include "f11vol.h"
#include "bitmap.h"

class index_file_bitmap : public bitmap
{
    friend class f11_file_indexf_sys;
    virtual dword_t free_value() { return 0; }
    virtual dword_t busy_value() { return 1; }
};

class storage_file_bitmap : public bitmap
{
    friend class f11_file_bitmap_sys;
    virtual dword_t free_value() { return 1; }
    virtual dword_t busy_value() { return 0; }
};

class f11_file_indexf_sys : public f11_file
{
private:
    index_file_bitmap	bmap;

    ods_dword_t* get_content_bmap() const { 
	ASSERT(p_content_window != NULL);
	
	// The bitmap starts at VBN hm2_w_ibmapvbn
	return (ods_dword_t*)
	       ((byte_t*)p_content_window->p_content + 
	       F11_BLOCK_SIZE * (dev->hblk.hm2_w_ibmapvbn - 1));
    }

public:
    void	init_bmap();

    dword_t	alloc_file_index();
    void	release_file_index(dword_t fid_num);
};

class f11_file_bitmap_sys : public f11_file
{
private:
    storage_file_bitmap	bmap;

    ods_dword_t* get_content_bmap() const { 
	ASSERT(p_content_window != NULL);
	
	// The bitmap starts at VBN 2
	return (ods_dword_t*)
	       ((byte_t*)p_content_window->p_content + F11_BLOCK_SIZE);
    }

public:
    void	init_bmap();

    ODS_SCB*	get_content_scb() const { 
	ASSERT(p_content_window != NULL);
		    
	// The SCB resides at VBN 1
	return (ODS_SCB*)(p_content_window->p_content); 
    }

    dword_t	get_free_clusters();
    dword_t	get_free_blocks();

    dword_t	lbn2cluster(lbn_t lbn) const {
	return (lbn / get_content_scb()->scb_w_cluster);
    }

    lbn_t	clusters2lbn(dword_t cluster) const {
	return (cluster * get_content_scb()->scb_w_cluster);
    }

    dword_t	alloc_clusters(dword_t n);
    dword_t	alloc_clusters_at(dword_t start_at, dword_t n);
    void	release_cluster(dword_t cluster);
};

#endif 	//__F11FILE_SYS_H__
