//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

// Files-11 Structure level 2 

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "definitions.h"
#include "f11asrt.h"
#include "devio.h"
#include "files11.h"
#include "f11liberr.h"
#include "f11file.h"
#include "f11file_dir.h"
#include "f11file_sys.h"
#include "f11vol.h"

static const ODS_FID	idx_fid(FID_INDEXF_NUMB, FID_INDEXF_SEQ, FID_RVN_INVOLCONT);

// Static Function Declarations
static int	filename_wild(const char*	file_spec);

static void	filenamescan(const char*	file_spec, 
			     dword_t*		p_length, 
			     dword_t*		p_version);
// ACP-QIO Functions
int f11_volume_acp::acp_sub_dirlookup(
			word_t		fib_w_nmctl_fib_v_wild,
			ODS_FID&	fib_w_fid,
			const ODS_FID&	fib_w_did,
			dword_t&	fib_l_wcc,
			const char*	p2,
			word_t*		p3,
			char*		p4 )
{
    f11_directory*	p_dir;
    ODS_dir_record*	dr;
    ODS_dir_record*	act_dr;
    ODS_dir_entry*	e;
    byte_t		ct2;

    // FIB$W_DID contains the file ID of the directory file. This
    // field must be a nonzero number.
    if (!fib_w_did.is_valid())
    {
	return SS_ERROR;
    }

    // Access the directory file with fid == fib_w_did
    p_dir = new f11_directory;

    // wsize == -1: map the entire file
    if (acp_sub_access(p_dir, -1, fib_w_did) == 0) 
    {
	delete p_dir;
	p_dir = NULL;
	return SS_ERROR;	//Cannot open
    }

    if (!p_dir->is_valid(NULL)) 
    {
	qio_deaccess(p_dir, fib_w_did);
	delete p_dir;
	p_dir = NULL;
	return SS_ERROR;	//Not a directory
    }

    // Extract pointer to dir record and dir entry index
    // from WCC.
    p_dir->decompose_wcc(fib_l_wcc, &dr, &e, &ct2);

    // Lookup for the next matching dir entry
    // starting from dir record dr and 
    // dir entry specified by ct2
    e = p_dir->f11_direntry_lookup(
		p2,
		fib_w_nmctl_fib_v_wild,
		dr, 
		ct2, 
		&dr,
		&act_dr);

    // Compose new WCC
    fib_l_wcc = p_dir->compose_wcc(dr, ct2);

    if (e == NULL) {
	qio_deaccess(p_dir, fib_w_did);
	delete p_dir;
	p_dir = NULL;
	return SS_ERROR;	// File not found
    } else {

	// The file ID of the file found is returned in FIB$W_FID.
	fib_w_fid = e->dir_w_fid;

	TRACE(("dirlookup successful: %-.*s;%d\n", act_dr->dir_b_namecount, act_dr->dir_t_name, (word_t)e->dir_w_version));
	
	// In addition, the resultant file name string is returned in
	// P4 and its length in P3.

	int t;
	if (p4 != NULL) {
	    t = sprintf(p4, "%-.*s;%d", act_dr->dir_b_namecount, act_dr->dir_t_name, (word_t)e->dir_w_version);
	}
	if (p3 != NULL)
	    *p3 = t;
    }

    // Finally, deaccess the directory file
    qio_deaccess(p_dir, fib_w_did);
    delete p_dir;
    p_dir = NULL;

    return SS_NORMAL;
}

int
f11_volume_acp::acp_sub_access(
			f11_file*	file,
			byte_t		fib_b_wsize,
			const ODS_FID&	fib_w_fid )
{
    lbn_t	lbn;
    file_hdr	*fl = NULL, *fl1 = NULL;
    ODS_FID	fid;

    TRACE(("f11_open_file: id 0%d\n", fib_w_fid.numb()));

    // Volume sets are not supported
    if(fib_w_fid.fid_b_rvn) {
	set_error(F11_VOLUME_SET);
	return SS_ERROR;
    }

    file->dev = this;

    // Load file header(s), including extension headers
    fid = fib_w_fid;

    do {
	// Get file header lbn
	if((lbn = get_hdr_lbn(fid.numb())) == INVALID_LBN) {
	    qio_deaccess(file, fid);
	    set_error(F11_INVALID_FH);
	    return SS_ERROR;
	}

	// Load file header
	fl = new file_hdr;

	TRACE(("qio_sub_access(): add header = %p\n", fl));
	if(!fl1) {
	    file->file_header = fl;
	    fl->prev = fl->next = NULL;
	} else {
	    fl1->next = fl;
	    fl->prev  = fl1;
	    fl->next = NULL;
	}

	if(f11_read_logical_blk(lbn, 1, &fl->fhdr) != F11_BLOCK_SIZE) 
	{
	    qio_deaccess(file, fid);
	    set_error(F11_DRIVE_ERR);
	    return SS_ERROR;
	}
	fl->fhdr_lbn = lbn;

	//  check file header, EOF BLK, HI BLK
	if(!fl->fhdr.is_file_header(&fid) ||
	   (fl->fhdr.get_hiblk() * F11_BLOCK_SIZE < 
		fl->fhdr.get_file_length())) 
	{
	    qio_deaccess(file, fid);
	    set_error(F11_INVALID_FH);
	    return SS_ERROR;
	}

	// Check seq number
	if(fl1 && fl1->fhdr.fh2_w_ext_fid.fid_w_seq !=  fid.fid_w_seq) 
	{
	    qio_deaccess(file, fid);
	    set_error(F11_INVALID_SEQ);
	    return SS_ERROR;
	}

	// Next file header 
	fid = fl->fhdr.fh2_w_ext_fid;
	fl1 = fl;

    } while(fid.numb());

    // Init file content window structure:

    // Special rule for the volume index file:
    // We do not load file headers, only home blocks and index file bitmap,
    // since we do not support caching of the blocks and load file
    // headers separately above!

    if (fib_w_fid == idx_fid) 
    {
	if (file->get_fhdr()->get_hiblk()) {

	    // The index file bitmap continues through VBN 
	    // HM2$W_CLUSTER * 4 + HM2$W_IBMAPSIZE, so load all blocks up to
	    // this value
	    dword_t load_vbns = this->hblk.hm2_w_cluster * 4 + (dword_t)this->hblk.hm2_w_ibmapsize;
	    file->p_content_window = new file_window(load_vbns);
	    if (file->f11_read_file_blk(
				0, load_vbns) != 
	       (int)load_vbns) 
	    {
		qio_deaccess(file, fid);
		return SS_ERROR;
	    }
	} else {
	    file->p_content_window = NULL;
	    return SS_ERROR;
	}
    } else 
    if (fib_b_wsize == (byte_t)0) {
	file->p_content_window = NULL;
    } else 
    if (file->get_fhdr()->is_file_contig() || fib_b_wsize == (byte_t)-1) {
	// Directory or contigous file
	// Make the cathedral window
	dword_t load_vbns = file->get_fhdr()->get_hiblk();
	if (load_vbns) {
	    file->p_content_window = new file_window(load_vbns);
	    if (file->f11_read_file_blk(
				0, load_vbns) != 
	       (int)load_vbns) 
	    {
		qio_deaccess(file, fid);
		return SS_ERROR;
	    }
	} else
	{
	    file->p_content_window = NULL;
	}
    } else {
	// Ordinary file, do not load contents now, do this later
	file->p_content_window = new file_window(fib_b_wsize);
    }

    TRACE(("f11_open_file: done \n"));

    return SS_NORMAL;
}

int
f11_volume_acp::acp_sub_create_file(
			ODS_FID&	fib_w_fid,
			const char*	p2
		    )
{
    lbn_t	lbn;
    file_hdr	*fl = NULL;

    // Allocate one of free file indexes for the file
    dword_t fid_num = p_indexf_sys->alloc_file_index();

    if (fid_num == -1)
    {
        // There is no free file header index
	// File system is full
        set_error(F11_IDX_FULL);
        return SS_ERROR;
    }

    TRACE(("INDEXF.SYS: bit %lu for fid_num = %lu allocated\n", fid_num - 1, fid_num));
    TRACE(("INDEXF.SYS ff_bit updated...\n"));
    p_indexf_sys->flag_content_modified = 1;

    // Create and fill the file header

    // Get file header lbn
    if((lbn = get_hdr_lbn(fid_num)) == INVALID_LBN) {
	set_error(F11_INVALID_FH);
	return SS_ERROR;
    }

    // Load file header
    fl = new file_hdr;
    fl->prev = fl->next = 0;

    if (f11_read_logical_blk(lbn, 1, &fl->fhdr) != F11_BLOCK_SIZE) 
    {
	set_error(F11_DRIVE_ERR);
	return SS_ERROR;
    }
    fl->fhdr_lbn = lbn;

    // Initialize empty content:
    fl->fhdr.fh2_b_idoffset = 40;
    fl->fhdr.fh2_b_mpoffset = 100;
    fl->fhdr.fh2_b_acoffset = 255;
    fl->fhdr.fh2_b_rsoffset = 255;

    fl->fhdr.fh2_w_seg_num = 0;		// This is not an extension header
    fl->fhdr.fh2_w_struclev = this->hblk.hm2_w_struclev;

    fl->fhdr.fh2_w_fid.fid_b_rvn = (byte_t)this->hblk.hm2_w_rvn;
    fl->fhdr.fh2_w_fid.fid_w_num = fid_num & 0x0000FFFF;
    fl->fhdr.fh2_w_fid.fid_b_nmx = (fid_num >> 16) & 0x00FF;

    // Update seq number
    fl->fhdr.fh2_w_fid.fid_w_seq = fl->fhdr.fh2_w_fid.fid_w_seq+1;

    fl->fhdr.fh2_w_ext_fid.fid_b_nmx = 0;
    fl->fhdr.fh2_w_ext_fid.fid_b_rvn = 0;
    fl->fhdr.fh2_w_ext_fid.fid_w_num = 0;
    fl->fhdr.fh2_w_ext_fid.fid_w_seq = 0;

    bzero((void*)&fl->fhdr.fh2_w_recattr, sizeof(ACP_FAT));	// Must be defined later to the actual value
    fl->fhdr.set_hiblk(0);
    fl->fhdr.set_file_length(0);
    bzero((void*)&fl->fhdr.fh2_l_filechar, sizeof(ODS_FCH));	// Must be defined later to the actual value

    fl->fhdr.fh2_pad0 = 0;
    fl->fhdr.fh2_b_acc_mode = 0;
    fl->fhdr.fh2_l_fileowner = this->hblk.hm2_l_volowner;
    fl->fhdr.fh2_w_fileprot = this->hblk.hm2_w_fileprot;

    // Must be defined later to the actual value:
    fl->fhdr.fh2_w_backlink.fid_b_nmx = 0;
    fl->fhdr.fh2_w_backlink.fid_b_rvn = 0;
    fl->fhdr.fh2_w_backlink.fid_w_num = 0;
    fl->fhdr.fh2_w_backlink.fid_w_seq = 0;

    fl->fhdr.fh2_b_journal = 0;
    fl->fhdr.fh2_b_ru_active = 0;

    fl->fhdr.fh2_pad1 = 0;

    fl->fhdr.fh2_l_highwater = 0;	// No highwater mark is maintained for the file

    fl->fhdr.fh2_pad2[0] = 0;
    fl->fhdr.fh2_pad2[1] = 0;
    fl->fhdr.fh2_pad2[2] = 0;
    fl->fhdr.fh2_pad2[3] = 0;

    fl->fhdr.fh2_r_class_prot[0] = 0;
    fl->fhdr.fh2_r_class_prot[1] = 0;
    fl->fhdr.fh2_r_class_prot[2] = 0;
    fl->fhdr.fh2_r_class_prot[3] = 0;
    fl->fhdr.fh2_r_class_prot[4] = 0;
    fl->fhdr.fh2_r_class_prot[5] = 0;
    fl->fhdr.fh2_r_class_prot[6] = 0;
    fl->fhdr.fh2_r_class_prot[7] = 0;
    fl->fhdr.fh2_r_class_prot[8] = 0;
    fl->fhdr.fh2_r_class_prot[9] = 0;

    ODS_FI2* ida = fl->fhdr.get_id_area();
    bzero(ida, fl->fhdr.fh2_b_mpoffset - fl->fhdr.fh2_b_idoffset);
    ida->string2fn(p2);

    ODS_FM2* mp  = fl->fhdr.get_mp_area();
    bzero(mp, fl->fhdr.fh2_b_acoffset - fl->fhdr.fh2_b_mpoffset);
    fl->fhdr.fh2_b_map_inuse = 0;	// No space allocated for content yet

    // The file ID of the file created is returned in FIB$W_FID
    fib_w_fid = fl->fhdr.fh2_w_fid;

    // Calculate new checksum for the file header block
    fl->fhdr.update_checksum();

    // and write it back to the volume
    if(f11_write_logical_blk(fl->fhdr_lbn, 1, &fl->fhdr) != F11_BLOCK_SIZE) 
    {
	delete fl;
	set_error(F11_DRIVE_ERR);
	return SS_ERROR;
    }

    delete fl;

    return SS_NORMAL;
}

int f11_volume_acp::acp_sub_extend(
			f11_file*	file,
			word_t		fib_w_exctl_fib_v_extend,
			word_t		fib_w_exctl_fib_v_alcon,
			word_t		fib_w_exctl_fib_v_alconb,
			word_t		fib_w_exctl_fib_v_filcon,
			word_t		fib_w_exctl_fib_v_aldef,
			dword_t		fib_l_exsz,
			dword_t&	fib_l_exvbn
		    )
{
    if (fib_w_exctl_fib_v_alcon || file->get_fhdr()->is_file_contig()) 
    {
	// Extend the file contigously
	dword_t allocated_clusters = blocks2clusters(file->get_fhdr()->get_hiblk(), this->hblk.hm2_w_cluster);
	dword_t new_placement = this->p_bitmap_sys->alloc_clusters(allocated_clusters + blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster));
	allocated_clusters += blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster);
	if (new_placement != (dword_t)-1 )
	{
	    // success, move the entire file (remap the file window and whole map area)
	    // to the new position:
	    TRACE(("Extend: moving the contiguos file to new placement!!!\n"));

	    // create new file window of the new (extended) size
	    file_window* new_file_window = new file_window(allocated_clusters*this->hblk.hm2_w_cluster);
	    for (unsigned int i = 0; i < new_file_window->block_count; i++)
	    {
		new_file_window->p_block_lbn[i] = new_placement*this->hblk.hm2_w_cluster + i;
	    }
	    new_file_window->start_vbn = 0;

	    if (file->p_content_window != NULL)
	    {
		ASSERT(file->p_content_window->block_count != 0);
		ASSERT(file->p_content_window->start_vbn == new_file_window->start_vbn);

		// move old content
		memcpy(new_file_window->p_content, file->p_content_window->p_content, file->p_content_window->block_count*F11_BLOCK_SIZE);
		for (unsigned int j = 0; j < file->p_content_window->block_count; j++)
		{
		    lbn_t lbn = file->p_content_window->p_block_lbn[j];
		    this->p_bitmap_sys->release_cluster(lbn / this->hblk.hm2_w_cluster);
		}
		delete file->p_content_window;
	    }
	    file->p_content_window = new_file_window;
	    file->flag_content_modified = 1;

	    // Update the file map area
	    ODS_FM2 *ida = file->get_fhdr()->get_mp_area();

	    // I decided to hardcode rp of FORMAT3 here.
	    ida->ret_ptr3.fm2_v_format = FM2_C_FORMAT3;
	    ida->ret_ptr3.fm2_l_lbn3[0] = (new_placement*this->hblk.hm2_w_cluster) & 0x00FF;
	    ida->ret_ptr3.fm2_l_lbn3[1] = (new_placement*this->hblk.hm2_w_cluster >> 8) & 0x00FF;
	    ida->ret_ptr3.fm2_l_lbn3[2] = (new_placement*this->hblk.hm2_w_cluster >> 16) & 0x00FF;
	    ida->ret_ptr3.fm2_l_lbn3[3] = (new_placement*this->hblk.hm2_w_cluster >> 24) & 0x00FF;
	    ida->ret_ptr3.fm2_w_lowcount[0] = ((allocated_clusters*this->hblk.hm2_w_cluster-1)) & 0x00FF;
	    ida->ret_ptr3.fm2_w_lowcount[1] = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 8) & 0x00FF;
	    ida->ret_ptr3.fm2_v_count2_1 = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 16) & 0x00FF;
	    ida->ret_ptr3.fm2_v_count2_2 = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 24) & 0x00FF;
	    file->get_fhdr()->fh2_b_map_inuse = ida->rp_getwords();

	    ASSERT(ida->rp_getcount() == allocated_clusters*this->hblk.hm2_w_cluster);
	    ASSERT(ida->rp_getlbn() == new_placement*this->hblk.hm2_w_cluster);

	    file->get_fhdr()->set_hiblk(allocated_clusters*this->hblk.hm2_w_cluster);
	    file->flag_header_modified = 1;

	    // sync file
	    file->f11_sync();
	    this->p_bitmap_sys->f11_sync();
	    return SS_NORMAL;

	} else
	{
	    return SS_ERROR;
	}
    } else {

	// Extend a non-contigous file
	file_hdr	*f, *f1;
	ODS_FH2*	fh;
	ODS_FM2  	*rp;
	ODS_FM2*	last_rp = NULL;
	ODS_FM2*	new_rp = NULL;
	dword_t		new_placement;
	dword_t		allocated_clusters;

	// First, find the last retrieval pointer in the map area.
	f = f1 = file->file_header;
	do {

	    f = f1;
	    ASSERT(f);

	    fh = &f->fhdr;
	    rp = fh->get_mp_area();

	    rp_iterator next_rp(rp, fh->fh2_b_map_inuse);
	    while ((rp = next_rp++) != NULL)
	    {
		if(rp->rp_isfmt0()) {
		    continue;
		}
		last_rp = rp;
	    }

	    //  next extension
	    f1 = f->next;
	} while(f1 != NULL);

	allocated_clusters = blocks2clusters(file->get_fhdr()->get_hiblk(), this->hblk.hm2_w_cluster);

	if (last_rp != NULL)
	{
	    lbn_t	lbn = last_rp->rp_getlbn();
	    dword_t cnt = last_rp->rp_getcount();
	    lbn_t	last_lbn = lbn + cnt;

	    new_rp = (ODS_FM2*)((word_t*)last_rp + last_rp->rp_getwords());

	    // Try to allocate blocks that could be connected to the last rp.
	    new_placement = this->p_bitmap_sys->alloc_clusters_at(last_lbn / this->hblk.hm2_w_cluster , blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster));
	    if (new_placement != (dword_t)-1 )
	    {
		// Success, 
		allocated_clusters += blocks2clusters(fib_l_exsz,this->hblk.hm2_w_cluster);

		// Modify the last rp to reflect newly allocated space
		last_rp->rp_setcount(last_rp->rp_getcount() + fib_l_exsz);

	    } else
	    {
		new_placement = this->p_bitmap_sys->alloc_clusters(blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster));
		if (new_placement != (dword_t)-1 )
		{
		    // Success, 
		    allocated_clusters += blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster);

		} else 
		{
		    return SS_ERROR;
		}

		// Create a new rp.
		if ((fh->fh2_b_map_inuse + 4/*i.e sizeof(ret_ptr3)/2*/) < 
		    fh->fh2_b_acoffset - fh->fh2_b_mpoffset)
		{
		    TRACE(("Extend: no rp detected, creating the first one.\n"));

		    // I decided to hardcode rp of FORMAT3 here.
		    new_rp->ret_ptr3.fm2_v_format = FM2_C_FORMAT3;
		    new_rp->ret_ptr3.fm2_l_lbn3[0] = (new_placement*this->hblk.hm2_w_cluster) & 0x00FF;
		    new_rp->ret_ptr3.fm2_l_lbn3[1] = (new_placement*this->hblk.hm2_w_cluster >> 8) & 0x00FF;
		    new_rp->ret_ptr3.fm2_l_lbn3[2] = (new_placement*this->hblk.hm2_w_cluster >> 16) & 0x00FF;
		    new_rp->ret_ptr3.fm2_l_lbn3[3] = (new_placement*this->hblk.hm2_w_cluster >> 24) & 0x00FF;
		    new_rp->ret_ptr3.fm2_w_lowcount[0] = ((allocated_clusters*this->hblk.hm2_w_cluster-1)) & 0x00FF;
		    new_rp->ret_ptr3.fm2_w_lowcount[1] = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 8) & 0x00FF;
		    new_rp->ret_ptr3.fm2_v_count2_1 = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 16) & 0x00FF;
		    new_rp->ret_ptr3.fm2_v_count2_2 = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 24) & 0x00FF;
		    fh->fh2_b_map_inuse = fh->fh2_b_map_inuse + new_rp->rp_getwords();
		}
		else
		{
		    // We need to create an extension header (not implemented yet)
		    return SS_ERROR;
		}
	    }

	}
	else
	{
	    last_rp = rp = fh->get_mp_area();
	    new_rp = rp;
	    new_placement = this->p_bitmap_sys->alloc_clusters(blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster));
		if (new_placement != (dword_t)-1 )
	    {
		// Success, 
		allocated_clusters += blocks2clusters(fib_l_exsz, this->hblk.hm2_w_cluster);

		TRACE(("Extend: no rp detected, creating the first one.\n"));

		// I decided to hardcode rp of FORMAT3 here.
		new_rp->ret_ptr3.fm2_v_format = FM2_C_FORMAT3;
		new_rp->ret_ptr3.fm2_l_lbn3[0] = (new_placement*this->hblk.hm2_w_cluster) & 0x00FF;
		new_rp->ret_ptr3.fm2_l_lbn3[1] = (new_placement*this->hblk.hm2_w_cluster >> 8) & 0x00FF;
		new_rp->ret_ptr3.fm2_l_lbn3[2] = (new_placement*this->hblk.hm2_w_cluster >> 16) & 0x00FF;
		new_rp->ret_ptr3.fm2_l_lbn3[3] = (new_placement*this->hblk.hm2_w_cluster >> 24) & 0x00FF;
		new_rp->ret_ptr3.fm2_w_lowcount[0] = ((allocated_clusters*this->hblk.hm2_w_cluster-1)) & 0x00FF;
		new_rp->ret_ptr3.fm2_w_lowcount[1] = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 8) & 0x00FF;
		new_rp->ret_ptr3.fm2_v_count2_1 = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 16) & 0x00FF;
		new_rp->ret_ptr3.fm2_v_count2_2 = ((allocated_clusters*this->hblk.hm2_w_cluster-1) >> 24) & 0x00FF;

		fh->fh2_b_map_inuse = new_rp->rp_getwords();
		ASSERT(new_rp->rp_getcount() == allocated_clusters*this->hblk.hm2_w_cluster);
		ASSERT(new_rp->rp_getlbn() == new_placement*this->hblk.hm2_w_cluster);

	    } else 
	    {
		return SS_ERROR;
	    }
	}

	file->get_fhdr()->set_hiblk(allocated_clusters*this->hblk.hm2_w_cluster);
	file->flag_header_modified = 1;

	// sync file
	file->f11_sync();
	this->p_bitmap_sys->f11_sync();
	return SS_NORMAL;
    }
    return SS_ERROR;
}

int f11_volume_acp::acp_sub_create_direntry(
			const ODS_FID&	fib_w_did,
			const char*	p2,
			const ODS_FID&	fib_w_fid,
			word_t		version,
			word_t*		p3,
			char*		p4
		    )
{
    return SS_ERROR;
}

int f11_volume_acp::acp_sub_create_dirrecord(
			const ODS_FID&	fib_w_did,
			const char*	p2,
			const ODS_FID&	fib_w_fid,
			word_t		version,
			word_t*		p3,
			char*		p4
		    )
{
    f11_directory*	p_dir;
    ODS_dir_record*	dr;
    ODS_dir_record*	prev_dr;
    ODS_dir_record*	new_dr;
    ODS_dir_entry*	e;

    // Access the directory file with fid == fib_w_did
    p_dir = new f11_directory;

    // wsize == -1: map the entire file
    if (acp_sub_access(p_dir, -1, fib_w_did) == 0) 
    {
	delete p_dir;
	p_dir = NULL;
	return SS_ERROR;	//Cannot open
    }

    if (!p_dir->is_valid(NULL)) 
    {
	qio_deaccess(p_dir, fib_w_did);
	delete p_dir;
	p_dir = NULL;
	return SS_ERROR;	//Not a directory
    }

    // Search for last dir record

    // Start with initial record...
    prev_dr = dr = (ODS_dir_record*)p_dir->p_content_window->p_content;

    for (;(dr != NULL); 
	 dr = p_dir->f11_get_next_record(dr))
    {
	if (dr->is_empty_record()) {
	    TRACE(("acp_sub_create_dirrecord: empty record, go to next record\n"));

	    // go to next record
	    continue;
	}
	prev_dr = dr;
    }

    if (prev_dr->is_empty_record()) {
	new_dr = prev_dr;
    }
    else {

	// prev_dr is the last dir record, skip to next, nonexisting
	new_dr = (ODS_dir_record*)((byte_t*)prev_dr + prev_dr->dir_w_size + 2);
    }

    if (((byte_t*)new_dr - (byte_t*)p_dir->p_content_window->p_content) / F11_BLOCK_SIZE > (int)(p_dir->get_fhdr()->get_hiblk()))
    {
	// Extend the directory file (not implemented yet)
	return SS_ERROR;
    }

    // Compose new dirrecord...
    new_dr->dir_w_verlimit = -1;
    new_dr->dir_b_flags.dir_v_type = DR_FID;
    new_dr->dir_b_namecount = strlen(p2);
    strncpy((char*)new_dr->dir_t_name, p2, new_dr->dir_b_namecount);

    // If the length of the name is an odd number, the name string is 
    // padded with a single null character
    if (new_dr->dir_b_namecount & 1)
    {
	new_dr->dir_t_name[new_dr->dir_b_namecount] = '\0';
    }
    e = (ODS_dir_entry*)((byte_t*)new_dr + 6 + new_dr->dir_b_namecount +
		    (new_dr->dir_b_namecount & 1));
    e->dir_w_fid = fib_w_fid;
    e->dir_w_version = version;

    new_dr->dir_w_size = 6 - 2 + new_dr->dir_b_namecount + 
	(new_dr->dir_b_namecount & 1) + sizeof(ODS_dir_entry);

    // Mark the record next after new as an skip record
    ((ODS_dir_record*)((byte_t*)(new_dr)+(new_dr->dir_w_size + 2)))->dir_w_size = -1;

    p_dir->flag_content_modified = 1;

    // In addition, the resultant file name string is returned in
    // P4 and its length in P3.

    int t;
    if (p4 != NULL) {
	t = sprintf(p4, "%-.*s;%d", new_dr->dir_b_namecount, new_dr->dir_t_name, (word_t)e->dir_w_version);
    }
    if (p3 != NULL)
	*p3 = t;

    // Finally, deaccess the directory file
    qio_deaccess(p_dir, fib_w_did);
    delete p_dir;
    p_dir = NULL;

    return SS_NORMAL;
}

int f11_volume_acp::acp_sub_truncate(
		    f11_file*	file,
		    dword_t&	fib_l_exsz,
		    dword_t	fib_l_exvbn )
{
    // Not implemented

    // Release the file blocks

    // Modify the file header

    return SS_ERROR;
}

// ACP-QIO Major Functions
// They are well documented in the ACP-QIO interface manual

int f11_volume_acp::qio_access(
			f11_file*	file,
			dword_t		io_m_create,
			dword_t		io_m_access,

			// for create
			word_t		fib_w_nmctl_fib_v_newver,
			word_t		fib_w_nmctl_fib_v_supersede,

			// for sub dirlookup:
			word_t		fib_w_nmctl_fib_v_wild,
			ODS_FID&	fib_w_fid,
			const ODS_FID&	fib_w_did,
			dword_t&	fib_l_wcc,

			// for sub access:
			byte_t		fib_b_wsize,

			// for extend
			word_t		fib_w_exctl_fib_v_extend,
			word_t		fib_w_exctl_fib_v_alcon,
			word_t		fib_w_exctl_fib_v_alconb,
			word_t		fib_w_exctl_fib_v_filcon,
			word_t		fib_w_exctl_fib_v_aldef,
			dword_t		fib_l_exsz,
			dword_t&	fib_l_exvbn,

			const char*	p2,
			word_t*		p3,
			char*		p4
		)
{
    dword_t sts = SS_NORMAL;

    // If a nonzero directory file ID is specified in FIB$W_DID,
    // the lookup function is performed.
    if (fib_w_did.is_valid()) {
	sts = acp_sub_dirlookup(fib_w_nmctl_fib_v_wild,
				fib_w_fid,
				fib_w_did,
				fib_l_wcc,
				p2,
				p3,
				p4
		    );
    }
    if (sts == SS_ERROR) { //NOSUCHFILE 

	// If the directory search fails with a 'file not found'
	// condition and the IO$M_CREATE function modifier is specified,
	// the function is reexecuted as CREATE.

	if (io_m_create)
	    return qio_create(file, 
				io_m_create, io_m_access, 
				0,
				fib_w_fid, fib_w_did,
				fib_w_nmctl_fib_v_newver,
				fib_w_nmctl_fib_v_supersede,
				fib_w_nmctl_fib_v_wild,
				fib_l_wcc,
				fib_b_wsize,
				fib_w_exctl_fib_v_extend,
				fib_w_exctl_fib_v_alcon,
				fib_w_exctl_fib_v_alconb,
				fib_w_exctl_fib_v_filcon,
				fib_w_exctl_fib_v_aldef,
				fib_l_exsz,
				fib_l_exvbn,
				p2,
				p3,
				p4
			);
	else 
	    return sts;
    }
    else {
    }

    // If IO$M_ACCESS is specified, the access subfunction 
    // is performed to open the file.
    if (io_m_access) {
	sts = acp_sub_access(file, fib_b_wsize, fib_w_fid);
    }

    return sts;
}

int f11_volume_acp::qio_create(f11_file* file,
			dword_t		io_m_create,
			dword_t		io_m_access,
			dword_t		io_m_delete,
			ODS_FID&	fib_w_fid,
			const ODS_FID&	fib_w_did,
			word_t		fib_w_nmctl_fib_v_newver,
			word_t		fib_w_nmctl_fib_v_supersede,

			// for sub dirlookup:
			word_t		fib_w_nmctl_fib_v_wild,
			dword_t&	fib_l_wcc,
			byte_t		fib_b_wsize,

			// for extend
			word_t		fib_w_exctl_fib_v_extend,
			word_t		fib_w_exctl_fib_v_alcon,
			word_t		fib_w_exctl_fib_v_alconb,
			word_t		fib_w_exctl_fib_v_filcon,
			word_t		fib_w_exctl_fib_v_aldef,
			dword_t		fib_l_exsz,
			dword_t&	fib_l_exvbn,

			const char*	p2,
			word_t*		p3,
			char*		p4
		)
{
    dword_t	sts = SS_NORMAL;
    int		has_wild = 0;
    int		has_record = 0;
    dword_t	length, version;

    char	fname_buf[256];
    ODS_FID	fid;
    dword_t	wcc = 0;

    has_wild = filename_wild(p2);

    if (has_wild || fib_w_nmctl_fib_v_wild)
	return sts = SS_ERROR;

    // Calculate input file_spec length and vesion
    filenamescan(p2, &length, &version);
    if(version == 0) version = 1;

    sprintf(fname_buf, "%.*s;*", length, p2);

    if (fib_w_did.is_valid()) {
	sts = acp_sub_dirlookup(fib_w_nmctl_fib_v_wild,
				fib_w_fid,
				fib_w_did,
				fib_l_wcc,
				p2,
				p3,
				p4
		    );
	if (sts == SS_NORMAL) {
	    set_error(F11_FILE_ALREADY_EXISTS);
	    return sts = SS_ERROR;  // Already exists
	}
    }

    if (fib_w_did.is_valid()) {
	sts = acp_sub_dirlookup(1, //fib_w_nmctl_fib_v_wild,
				fid,
				fib_w_did,
				wcc,
				fname_buf,
				NULL,
				NULL
		    );
    }

    // If the modifier IO$M_CREATE is specified, a file is created.
    // The file ID of the file created is returned in FIB$W_FID.
    if (io_m_create) {
	sts = acp_sub_create_file(fib_w_fid, p2);
	if (sts == SS_ERROR)
	    return sts;
    }

    // If the modifier IO$M_DELETE is specified, the file is marked
    // for deletion.

    // If a nonzero directory file ID is specified in FIB$W_DID,
    // a directory entry is created.
    sprintf(fname_buf, "%.*s", length, p2);

    if (fib_w_did.is_valid()) {
	if (has_record) {
	    sts = acp_sub_create_direntry(fib_w_did, fname_buf, fib_w_fid, version, p3, p4);
	    if (sts == SS_ERROR)
		return sts;
	} else
	{
	    sts = acp_sub_create_dirrecord(fib_w_did, fname_buf, fib_w_fid, version, p3, p4);
	    if (sts == SS_ERROR)
		return sts;
	}
    }

    // If IO$M_ACCESS is specified, the access subfunction 
    // is performed to open the file.
    if (io_m_access) {
	sts = acp_sub_access(file, fib_b_wsize, fib_w_fid);
	if (sts == SS_ERROR) {
	    return sts;
	}
    }

    // If the modifier IO$M_DELETE is specified, the file is marked
    // for deletion.
    if (io_m_delete) {

	// Mark it for deletion
	file->get_fhdr()->fh2_l_filechar.u.s.fch_v_markdel = 1;
	file->flag_header_modified = 1;
    }
    if (sts == SS_ERROR)
	return sts;

    // If the extend bit FIB$V_EXTEND is specified in the FIB,
    // the extend subfunction is performed.
    if (fib_w_exctl_fib_v_extend) {
	sts = acp_sub_extend(file,
			fib_w_exctl_fib_v_extend,
			fib_w_exctl_fib_v_alcon,
			fib_w_exctl_fib_v_alconb,
			fib_w_exctl_fib_v_filcon,
			fib_w_exctl_fib_v_aldef,
			fib_l_exsz,
			fib_l_exvbn
	    );
    }

    // Now only implemented: the backlink (fib_w_did) is entered in the header
    if (fib_w_did.is_valid())
    {
	file->set_backlink(fib_w_did);
    }

    return sts;
}

int f11_volume_acp::qio_deaccess(
			f11_file*	f,
			const ODS_FID&	fib_w_fid
		)
{
    dword_t	sts = SS_NORMAL;
    file_hdr	*fl, *fl1;

    TRACE(("qio_deaccess(): fid_num = %u\n", fib_w_fid.numb()));
    ASSERT(f);

    // Finally, the file is closed.
    f->f11_sync();

    // File is marked for deletion. Delete it on deaccess
    if (f->get_fhdr()->fh2_l_filechar.u.s.fch_v_markdel == 1)
    {
	TRACE(("qio_deaccess(): deleting file fid_num = %u\n", fib_w_fid.numb()));
	f->delete_file();
	this->f11_sync();
    }

    // Free list of file headers
    fl = f->file_header;
    while(fl) {
	TRACE(("qio_deaccess(): free header = %p\n", fl));
	fl1 = fl->next;
	delete fl;
	fl  = fl1;
    }

    // Free file contents
    if(f->p_content_window != NULL) {
	TRACE(("qio_deaccess(): free content = %p\n", f->p_content_window));
	delete f->p_content_window;
	f->p_content_window = NULL;
    }

    TRACE(("qio_deaccess(): done.\n"));
    return sts;
}

int f11_volume_acp::qio_delete(
			dword_t		io_m_delete,

			// for sub dirlookup:
			word_t		fib_w_nmctl_fib_v_wild,
			ODS_FID&	fib_w_fid,
			const ODS_FID&	fib_w_did,
			dword_t&	fib_l_wcc,

			const char*	p2,
			word_t*		p3,
			char*		p4
		)
{
    dword_t	sts = SS_NORMAL;

    // If a nonzero directory ID is specified in FIB$W_DID, the
    // directory lookup is performed.
    if (fib_w_did.is_valid())
    {
	f11_directory*	p_dir;
	ODS_dir_record*	dr;
	ODS_dir_record*	act_dr;
	ODS_dir_entry*	e;
	byte_t		ct2;

	p_dir = new f11_directory;

	// wsize == -1: map the entire file
	if (acp_sub_access(p_dir, -1, fib_w_did) == 0) 
	{
	    delete p_dir;
	    p_dir = NULL;
	    return SS_ERROR;    //Cannot open
	}

	if (!p_dir->is_valid(NULL)) 
	{
	    qio_deaccess(p_dir, fib_w_did);
	    delete p_dir;
	    p_dir = NULL;
	    return SS_ERROR;    //Not a directory
	}

	p_dir->decompose_wcc(fib_l_wcc, &dr, &e, &ct2);

	e = p_dir->f11_direntry_lookup(
		    p2,
		    fib_w_nmctl_fib_v_wild,
		    dr, 
		    ct2, 
		    &dr,
		    &act_dr);

	fib_l_wcc = p_dir->compose_wcc(dr, ct2);

	if (e == NULL) {
	    qio_deaccess(p_dir, fib_w_did);
	    delete p_dir;
	    p_dir = NULL;
	    return SS_ERROR;    // File not found
	} else {
	    
	    // The file ID of the file found is returned in FIB$W_FID.
	    fib_w_fid = e->dir_w_fid;

	    // In addition, the resultant file name string is returned in
	    // P4 and its length in P3.
	    TRACE(("qio_delete: %-.*s;%d\n", act_dr->dir_b_namecount, act_dr->dir_t_name, (word_t)e->dir_w_version));
	    int t;
	    if (p4 != NULL) {
		t = sprintf(p4, "%-.*s;%d", act_dr->dir_b_namecount, act_dr->dir_t_name, (word_t)e->dir_w_version);
	    }
	    if (p3 != NULL)
		*p3 = t;
	}

	// The file name located is removed from the directory
	if (sts == SS_NORMAL) {
	    p_dir->remove_dir_entry(act_dr, e);
	    p_dir->flag_content_modified = 1;
	}

	qio_deaccess(p_dir, fib_w_did);
	delete p_dir;
	p_dir = NULL;
    }

    // If the function modifier IO$M_DELETE is specified, the file
    // is marked for deletion.
    if (io_m_delete) 
    {
	// Access the file
	f11_file* p_file = new f11_file;

	// wsize == 0: do not map the file content
	if (acp_sub_access(p_file, 0, fib_w_fid) == 0) 
	{
	    delete p_file;
	    p_file = NULL;
	    return SS_ERROR;    //Cannot open
	}

	// Mark it for deletion
	p_file->get_fhdr()->fh2_l_filechar.u.s.fch_v_markdel = 1;
	p_file->flag_header_modified = 1;

	// And delete it on deaccess explicitly
	qio_deaccess(p_file, fib_w_fid);
	delete p_file;
	p_file = NULL;
    }

    return sts;

}

// Static functions

static int filename_wild(const char* file_spec)
{
    char*	b1;
    char*	b2;
    int wild	= 0;

    b1 = strchr(file_spec, '*');
    b2 = strchr(file_spec, '%');

    if ((b1 != 0) || (b2 != 0)) {
	wild = 1;
    }
    return wild;
}

static void filenamescan(const char* file_spec, dword_t* p_length, dword_t* p_version)
{
    char*	b;
    char*	p_e;

    b = strchr(file_spec, ';');
    if(b != 0) {
	*p_length  = (dword_t)(b - file_spec);
	*p_version = (dword_t)strtol(b + 1, &p_e, 10);
    } else {
	*p_length  = (dword_t)strlen(file_spec);
	*p_version = (dword_t)0;
    }
}
