//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

// Unix implementation

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>

#include "definitions.h"
#include "fileio.h"
#include "f11asrt.h"
#include "files11.h"

HANDLE fileio_open_file(const char* fn, int mode)
{
    int oflag;
    oflag = (mode & FILEIO_RO) ? O_RDONLY : (O_RDWR | O_TRUNC | O_CREAT);
    return (HANDLE)open(fn, oflag, 0666);
}

int fileio_read_file(HANDLE h, void* buf, int length)
{
    return read(h, buf, length);
}

int fileio_write_file(HANDLE h, void* buf, int length)
{
    return write(h, buf, length);
}

void	fileio_close_file(HANDLE h)
{
    close(h);
}

dword_t fileio_file_size(const char* fs)
{
    struct stat st;

    ASSERT(fs);
    if(stat(fs, &st) == -1) return -1;
    return st.st_size;
}

int fileio_file_chdir(const char* dir)
{
    ASSERT(dir);
    if(chdir(dir) == -1) return -1;
    return 0;
}

char* fileio_file_getwd(char* dir, int size)
{
    ASSERT(dir);
    return getcwd(dir, size); 
}

char* fileio_absolute_file_name(const char* rel, char* absolute, int len)
{
    ASSERT(rel && absolute);
    getcwd(absolute, len);
    
    if(rel[0] == '/' || 
       (len < strlen(absolute) + strlen(rel) + 2)) {
	    strncpy(absolute, rel, len);
    } else {
	sprintf(absolute + strlen(absolute), "/%s", rel);
    }
    return absolute;
}
