//
// Copyright (C) 2000 Software Resources International SA
//
// This code is provided on an 'as is' basis.
// Software Resources International SA carries no responsibility for
// the application of this code in any way or form.
//

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include "definitions.h"
#include "devio.h"
#include "f11asrt.h"
#include "ft11.h"

int cmd_done = 0;

static ft11_cmd cmd[] = {
    {"?",		1, ft11_help_cmd},
    {"ascii",		3, ft11_ascii_cmd},
    {"bin",		2, ft11_bin_cmd},
    {"bye",		2, ft11_bye_cmd},
    {"cd",		2, ft11_cd_cmd},
    {"cdup",		3, ft11_cdup_cmd},
    {"close",		2, ft11_close_cmd},
    {"debug",		3, ft11_debug_cmd},
    {"dir",		3, ft11_dir_cmd},
    {"dir/date",	5, ft11_dir_date_cmd},
    {"dir/file_id",	5, ft11_dir_fileid_cmd},
    {"dir/prot",	5, ft11_dir_prot_cmd},
    {"dir/size",	5, ft11_dir_size_cmd},
    {"disconnect",	3, ft11_close_cmd},
    {"exit",		2, ft11_bye_cmd},
    {"get",		3, ft11_get_cmd},
    {"help",		1, ft11_help_cmd},
    {"lcd",		2, ft11_lcd_cmd},
    {"lo",		2, ft11_bye_cmd},
    {"ls",		2, ft11_ls_cmd},
    {"mode",		2, ft11_mode_cmd},
    {"open",		2, ft11_open_cmd},
    {"put",		3, ft11_put_cmd},
    {"pwd",		3, ft11_pwd_cmd},
    {"quit",		1, ft11_bye_cmd},
    {"recv",		2, ft11_get_cmd},
    {"show_dev",	2, ft11_show_dev_cmd},
    {"show_dev/full",  13, ft11_show_dev_full_cmd},
    {"show_def",	8, ft11_show_def_cmd},
    {"status",		2, ft11_show_dev_full_cmd},
    {"vi",		2, ft11_vi_cmd}
};

static char* skip_spaces(char *p) 
{
    while(*p == ' ' || *p == '\t' || *p == '\n') p++;
    return p;
}

static char* skip_word(char* p)
{
    while(*p != ' ' && *p != '\t' && *p != '\n') p++;
    return p;
}

static char  cmd_str[MAX_COMMAND_LENGTH + 1];
static char* cmd_words[MAX_COMMAND_LENGTH/2];

int main(int argc, char* argv)
{
    int		num_words;
    FILE*	input = stdin;

    F11_Install();

    //	Parse command line options

    //	Process commands
    while(!cmd_done){
	memset(cmd_str, '\0', MAX_COMMAND_LENGTH + 1);
	printf("$ ");
	fgets(cmd_str, MAX_COMMAND_LENGTH, input);

	if (feof(input)) {
	    ft11_bye_cmd(0, 0);
	    break;
	}
	if (!strlen(cmd_str)) 
	    continue;

	if (*skip_spaces(cmd_str) == '!') {
	    cmd_str[strlen(cmd_str) - 1] = '\0';
	    system(skip_spaces(cmd_str) + 1);
	    printf("\n");
	    continue;
	}

	//  Splitting command string to array of the words
	num_words = cmd_split(cmd_str, cmd_words);
	if (!num_words) 
	    continue;

	//  Matching command and call cammand handler
	cmd_match(num_words, cmd_words);
    }
    printf("\n");
    return 0;
}

int  cmd_split(char* cmd_str, char* cmd_words[])
{
    char*	p = cmd_str;
    int 	i = 0;

    p = skip_spaces(p);
    while(strlen(p)) {
	cmd_words[i] = p;
	p = skip_word(p);
	*p++ = '\0';
	p = skip_spaces(p);
	i++;
    }
    return i;
}

void cmd_match(int num_words, char* cmd_words[])
{
    int i;
    if (!num_words) 
	return;

    for (i = 0; i < sizeof(cmd)/sizeof(ft11_cmd); i++) {
	if (strlen(cmd_words[0]) < (dword_t)cmd[i].min_chars) 
	    continue;
	if (strncmp(cmd_words[0], cmd[i].cmd, strlen(cmd_words[0])) == 0) {
	    num_words--;
	    cmd[i].f(num_words, num_words ? &cmd_words[1] : NULL);
	    break;
	}
    }
    if (i == sizeof(cmd)/sizeof(ft11_cmd)) {
	f11_error("?Invalid command");
    }
}


dword_t ft11_options(int argc, char* args[], int num_opt, char* opt[])
{
    int	val = 0, i, j;

    ASSERT(argc >= 0);

    for (i = 0; i < argc; i++ ) {
	ASSERT(args && args[i]);

	for (j = 0; j < num_opt; j++) {
	    ASSERT(opt && opt[j]);

	    if (strcmp(opt[j], args[i]) == 0) {
		val |= 1 << j;
	    }
	}
    }
    return val;
}

char* ft11_next_arg(int argc, char* args[], dword_t* cur)
{
    dword_t 	i;

    ASSERT(cur && *cur > 0);

    for (i = *cur; i < (dword_t)argc; i++, (*cur)++) {
	ASSERT(args && args[i]);
	if (args[i][0] != '-') 
	    return args[i];
    }

    return 0;
}

